﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace CompositePlasticBag
{
    public class StuffContainer : IStuff, IEnumerable<IStuff>
    {
        private readonly IList<IStuff> _stuffInContainer = new List<IStuff>();

        /// <summary>
        /// The description of this object
        /// </summary>
        public String Description { get; set; }

        /// <summary>
        /// The maximum weight that this container can carry
        /// </summary>
        public int MaxWeight { get; set; }

        /// <summary>
        /// true if this container is empty, false otherwise
        /// </summary>
        public bool IsEmpty { get { return _stuffInContainer.Count == 0; } }

        /// <summary>
        /// The number of OrdinarStuff in this container, directly + indirectly
        /// </summary>
        public int Count
        {
            get
            {
                //return _stuffInContainer.Count;
                int howMany = 0;
                foreach (IStuff stuff in _stuffInContainer)
                {
                    StuffContainer container = stuff as StuffContainer;
                    if (container != null)
                    {
                        StuffContainer cont = container;
                        howMany = howMany + cont.Count;
                    }
                    else
                    {
                        howMany = howMany + 1;
                    }
                }
                return howMany;
            }
        }

        /// <summary>
        /// Adds another piece of stuff (ordinary or container) to this container.
        /// If the new piece of stuff makes the weight exceed the MaxWeight, then the add is rejected, and false is returned.
        /// </summary>
        /// <param name="stuff">the stuff to be added</param>
        /// <returns>true if the stuff was added, false otherwise</returns>
        public bool Add(IStuff stuff)
        {
            if (Weight + stuff.Weight <= MaxWeight)
            {
                _stuffInContainer.Add(stuff);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Removes the specified piece of stuff (ordinary or container) from this container.
        /// </summary>
        /// <param name="stuff">the stuff to be removed</param>
        /// <returns>true if the stuff was in the container and is now removed, otherwise false</returns>
        public bool Remove(IStuff stuff)
        {
            return _stuffInContainer.Remove(stuff);
        }

        /// <summary>
        /// Returns true if the specified stuff (ordinary or container) is contained (directly or indirectly) in this container.
        /// </summary>
        /// <param name="stuff">the stuff to look for</param>
        /// <returns>true if the specified stuff (ordinary or container) is contained (directly or indirectly) in this container.</returns>
        public bool Contains(IStuff stuff)
        {
            foreach (IStuff st in _stuffInContainer)
            {
                if (st.Equals(stuff)) return true;
                StuffContainer container = st as StuffContainer;
                if (container != null)
                {
                    if (container.Contains(stuff))
                        return true;
                }
            }
            return false;
        }

        /// <summary>
        /// The accumulated weight of the ordinary stuff in this container, and in possible sub-containers
        /// </summary>
        public int Weight
        {
            get { return _stuffInContainer.Sum(stuff => stuff.Weight); }
        }

        /// <summary>
        /// Returns an enumerator over this container
        /// </summary>
        /// <returns>an enumerator over this container</returns>
        public IEnumerator<IStuff> GetEnumerator()
        {
            return _stuffInContainer.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        /// <summary>
        /// Returns a string representation of this container
        /// </summary>
        /// <returns>a string representation of this container</returns>
        public override string ToString()
        {
            return string.Format("Description: {0}, MaxWeight: {1}, contains: {2}", Description, MaxWeight, String.Join(", ", _stuffInContainer));
        }
    }
}
